import { requireAuthentication, t } from "./base"
import { prismaClient } from "../prisma"
import z from "zod"
import { observable } from "@trpc/server/observable"
import type { GameAction } from "../../shared/game/actions"
import { isDev } from "../isDev"
import { gameRouter } from "./game"
import { nanoid } from "nanoid/async"
import { createGame, getGameByLobbyCode } from "../game"
import type { GameEvent } from "../../shared/game/events"

export const appRouter = t.router({
	game: gameRouter,

	login: t.procedure
		.input(z.object({
			name: z.string().min(1).max(20)
		}))
		.mutation(async ({ input, ctx }) => {
			if (ctx.user !== null) await prismaClient.user.delete({ where: { id: ctx.user.id } })

			const token = await nanoid(60)
			const user = await prismaClient.user.create({
				data: {
					name: input.name,
					token
				},
				select: {
					id: true
				}
			})

			ctx.res!.cookie("token", token, {
				maxAge: 60 * 60 * 24 * 365,
				httpOnly: true,
				secure: !isDev,
				sameSite: "strict"
			})

			return {
				id: user.id
			}
		}),

	getSelf: t.procedure
		.query(async ({ ctx }) => {
			if (ctx.user === null) return { user: null }

			await prismaClient.user.update({
				where: { id: ctx.user.id },
				data: {
					lastActivityDate: new Date()
				}
			})

			return {
				user: await prismaClient.user.findUnique({
					where: { id: ctx.user.id },
					select: {
						id: true,
						name: true
					}
				})
			}
		}),

	createGame: t.procedure
		.use(requireAuthentication)
		.mutation(async ({}) => {
			const game = createGame()

			return {
				lobbyCode: game.lobbyCode
			}
		}),

	join: t.procedure
		.use(requireAuthentication)
		.input(z.object({
			lobbyCode: z.string().nonempty()
		}))
		.subscription(async ({ input, ctx }) => {
			const game = await getGameByLobbyCode(input.lobbyCode)
			if (game === null) throw new Error("There is no game with this code.")

			await game.addPlayer(ctx.user.id, ctx.user.name)
			
			return observable<GameEvent>(emit => {
				const handleBroadcastAction = (action: GameAction) => emit.next({ type: "action", action })
				const handleDestroyed = () => setTimeout(() => emit.complete(), 500)

				const handlePrivateAction = (playerId: string, action: GameAction) => {
					if (playerId === ctx.user.id) emit.next({ type: "action", action })
				}

				game.on("broadcast_action", handleBroadcastAction)
				game.on("private_action", handlePrivateAction)
				game.on("destroyed", handleDestroyed)

				game.sendAllOldActionsTo(ctx.user.id)

				return () => {
					game.off("broadcast_action", handleBroadcastAction)
					game.off("private_action", handlePrivateAction)
					game.off("destroyed", handleDestroyed)
				}
			})
		})
})

export type AppRouter = typeof appRouter