import type { SpecialCardType } from "./cards"
import type { GameAction } from "./actions"
import { produce } from "immer"
import { specialCardTypes } from "./cards"
import { cloneDeep, tail, without } from "lodash-es"

type SpecialCardCountByType = Record<SpecialCardType, number>

export interface GameStateNumberCard {
	number: number
	isCovert: boolean
}

export const getNumberCardsSum = (cards: Readonly<GameStateNumberCard[]>) => cards.reduce((acc, card) => acc + card.number, 0)

export interface GameStatePlayer {
	id: string
	name: string
	numberCards: GameStateNumberCard[]
	specialCardCountByType: SpecialCardCountByType
	stayed: boolean
	nextRoundCovert: boolean
}

export interface GameState {
	phase: "pre-start" | "running" | "end"
	players: GameStatePlayer[]
	activeSpecialCardCountByType: SpecialCardCountByType
	activePlayerId: string
	targetSum: number
	numberCardsStack: number[]  // if redacted: contains more cards than there actually are
	actualNumberCardsStackSize: number  // the length of numberCardsStack, smaller if redacted
	winnerIds: string[] | null
}

const ALL_ZERO_SPECIAL_CARD_COUNTS: SpecialCardCountByType = Object.fromEntries(specialCardTypes.map(t => [t, 0])) as SpecialCardCountByType

const UNINITIALIZED_GAME_STATE: GameState = {
	phase: "pre-start",
	players: [],
	activePlayerId: "",
	targetSum: 0,
	activeSpecialCardCountByType: ALL_ZERO_SPECIAL_CARD_COUNTS,
	numberCardsStack: [],
	actualNumberCardsStackSize: 0,
	winnerIds: null
}

export const getUninitializedGameState = () => cloneDeep(UNINITIALIZED_GAME_STATE)
export const getFullNumbersCardStack = () => tail([...Array(12).keys()])

export const produceNewState = (oldState: GameState, action: GameAction) => produce(oldState, state => {
	const activateNextPlayer = () => {
		const activePlayerIndex = state.players.findIndex(p => p.id === state.activePlayerId)
		state.activePlayerId = state.players[(activePlayerIndex + 1) % state.players.length].id
	}

	switch (action.type) {
		case "join":
			state.players.push({
				id: action.initiatingPlayerId,
				name: action.name,
				numberCards: [],
				specialCardCountByType: ALL_ZERO_SPECIAL_CARD_COUNTS,
				stayed: false,
				nextRoundCovert: false
			})

			break

		case "start":
			state.phase = "running"
			state.activePlayerId = state.players[0].id
			state.targetSum = action.targetSum
			state.activeSpecialCardCountByType = ALL_ZERO_SPECIAL_CARD_COUNTS
			state.numberCardsStack = getFullNumbersCardStack()
			state.actualNumberCardsStackSize = state.numberCardsStack.length

			break

		case "deal-number":
			const p1 = state.players.find(p => p.id === action.toPlayerId)!
			p1.numberCards.push({
				number: action.number,
				isCovert: action.isCovert
			})

			state.numberCardsStack = without(state.numberCardsStack, action.number)
			state.actualNumberCardsStackSize--

			if (state.actualNumberCardsStackSize === 0) {
				state.numberCardsStack = getFullNumbersCardStack()
				state.actualNumberCardsStackSize = state.numberCardsStack.length
			}

			break

		case "deal-special":
			const p2 = state.players.find(p => p.id === action.toPlayerId)!
			p2.specialCardCountByType[action.cardType]++
			break

		case "hit":
			activateNextPlayer()
			const p4 = state.players.find(p => p.id === action.initiatingPlayerId)!
			p4.stayed = false
			break

		case "stay":
			activateNextPlayer()
			const p5 = state.players.find(p => p.id === action.initiatingPlayerId)!
			p5.stayed = true
			break

		case "use-special":
			const p3 = state.players.find(p => p.id === action.initiatingPlayerId)!
			applySpecialCardUsage(state, action.cardType, p3)
			break

		case "end":
			state.phase = "end"
			state.winnerIds = action.winnerIds
			for (let player of state.players) {
				player.numberCards = action.cardsByPlayerId[player.id].map((number, index) => ({ number, isCovert: player.numberCards[index].isCovert }))
			}
	}
})

function applySpecialCardUsage(state: GameState, type: SpecialCardType, player: GameStatePlayer) {
	player.specialCardCountByType[type]--

	switch (type) {
		case "add-2-opponent":
			// nothing
			break

		case "destroy-last-opponent-special":
			// TODO
			break

		case "double-draw":
			// nothing
			break

		case "increase-target-by-2":
			state.targetSum += 2
			break

		case "next-round-covert":
			player.nextRoundCovert = true
			break

		case "return-last-opponent":
			// TODO
			break

		case "return-last-own":
			// TODO
			break
	}
}